/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.structure;

import jalview.datamodel.AlignmentAnnotation;
import jalview.datamodel.SequenceI;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

public class StructureMapping
{
  public static final int UNASSIGNED = -1;

  public static final int PDB_RES_NUM_INDEX = 0;

  public static final int PDB_ATOM_NUM_INDEX = 1;

  /**
   * Space character constant, for consistent representation when no chain
   * specified
   */
  public static String NO_CHAIN = " ";

  String mappingDetails;

  SequenceI sequence;

  String pdbfile;

  String pdbid;

  String pdbchain;

  // Mapping key is residue index while value is an array containing PDB resNum,
  // and atomNo
  HashMap<Integer, int[]> mapping;

  /**
   * Constructor
   * 
   * @param seq
   * @param pdbfile
   * @param pdbid
   * @param chain
   * @param mapping
   *          a map from sequence to two values, { resNo, atomNo } in the
   *          structure
   * @param mappingDetails
   */
  public StructureMapping(SequenceI seq, String pdbfile, String pdbid,
          String chain, HashMap<Integer, int[]> mapping,
          String mappingDetails)
  {
    sequence = seq;
    this.pdbfile = pdbfile;
    this.pdbid = pdbid;
    this.pdbchain = chain;
    this.mapping = mapping;
    this.mappingDetails = mappingDetails;
  }

  public SequenceI getSequence()
  {
    return sequence;
  }

  public String getChain()
  {
    return pdbchain;
  }

  public String getPdbId()
  {
    return pdbid;
  }

  /**
   * Answers the structure atom number mapped to the given sequence position, or
   * -1 if no mapping
   * 
   * @param seqpos
   * @return
   */
  public int getAtomNum(int seqpos)
  {
    int[] resNumAtomMap = mapping.get(seqpos);
    return (resNumAtomMap == null ? UNASSIGNED
            : resNumAtomMap[PDB_ATOM_NUM_INDEX]);
  }

  /**
   * Answers the structure residue number mapped to the given sequence position,
   * or -1 if no mapping
   * 
   * @param seqpos
   * @return
   */
  public int getPDBResNum(int seqpos)
  {
    int[] resNumAtomMap = mapping.get(seqpos);
    return (resNumAtomMap == null ? UNASSIGNED
            : resNumAtomMap[PDB_RES_NUM_INDEX]);
  }

  /**
   * Returns a (possibly empty) list of [start, end] residue positions in the
   * mapped structure, corresponding to the given range of sequence positions
   * 
   * @param fromSeqPos
   * @param toSeqPos
   * @return
   */
  public List<int[]> getPDBResNumRanges(int fromSeqPos, int toSeqPos)
  {
    List<int[]> result = new ArrayList<int[]>();
    int startRes = -1;
    int endRes = -1;

    for (int i = fromSeqPos; i <= toSeqPos; i++)
    {
      int resNo = getPDBResNum(i);
      if (resNo == UNASSIGNED)
      {
        continue; // no mapping from this sequence position
      }
      if (startRes == -1)
      {
        startRes = resNo;
        endRes = resNo;
      }
      if (resNo >= startRes && resNo <= endRes)
      {
        // within the current range - no change
        continue;
      }
      if (resNo == startRes - 1)
      {
        // extend beginning of current range
        startRes--;
        continue;
      }
      if (resNo == endRes + 1)
      {
        // extend end of current range
        endRes++;
        continue;
      }

      /*
       * resNo is not within or contiguous with last range,
       * so write out the last range
       */
      result.add(new int[] { startRes, endRes });
      startRes = resNo;
      endRes = resNo;
    }

    /*
     * and add the last range
     */
    if (startRes != -1)
    {
      result.add(new int[] { startRes, endRes });
    }

    return result;
  }

  /**
   * Answers the sequence position mapped to the given structure residue number,
   * or -1 if no mapping is found
   * 
   * @param pdbResNum
   * @return
   */
  public int getSeqPos(int pdbResNum)
  {
    for (Entry<Integer, int[]> map : mapping.entrySet())
    {
      if (pdbResNum == map.getValue()[PDB_RES_NUM_INDEX])
      {
        return map.getKey();
      }
    }
    return UNASSIGNED;
  }

  /**
   * transfer a copy of an alignment annotation row in the PDB chain coordinate
   * system onto the mapped sequence
   * 
   * @param ana
   * @return the copy that was remapped to the mapped sequence
   * @note this method will create a copy and add it to the dataset sequence for
   *       the mapped sequence as well as the mapped sequence (if it is not a
   *       dataset sequence).
   */
  public AlignmentAnnotation transfer(AlignmentAnnotation ana)
  {
    AlignmentAnnotation ala_copy = new AlignmentAnnotation(ana);
    SequenceI ds = sequence;
    while (ds.getDatasetSequence() != null)
    {
      ds = ds.getDatasetSequence();
    }
    // need to relocate annotation from pdb coordinates to local sequence
    // -1,-1 doesn't look at pdbresnum but fails to remap sequence positions...

    ala_copy.remap(ds, mapping, -1, -1, 0);
    ds.addAlignmentAnnotation(ala_copy);
    if (ds != sequence)
    {
      // mapping wasn't to an original dataset sequence, so we make a copy on
      // the mapped sequence too
      ala_copy = new AlignmentAnnotation(ala_copy);
      sequence.addAlignmentAnnotation(ala_copy);
    }
    return ala_copy;
  }

  public String getMappingDetailsOutput()
  {
    return mappingDetails;
  }

}
